pub mod cpu;
mod instructions;
mod keyboard;
pub mod memory;
mod platform;
pub mod video;

use crate::cpu::Cpu;
use crate::keyboard::Keyboard;
use crate::memory::Mem;
use crate::video::Renderer;

use cpu::CpuController;
use memory::MemHandle;
use minifb::Window;

#[cfg(not(target_arch = "wasm32"))]
use platform::native as imp;
#[cfg(target_arch = "wasm32")]
use platform::wasm as imp;

pub struct GeorgeEmu(imp::GeorgeEmu);

impl GeorgeEmu {
    pub fn builder() -> GeorgeEmuBuilder<NoRom, NoWindow> {
        GeorgeEmuBuilder::new()
    }

    pub fn draw(&mut self) {
        self.0.draw()
    }

    pub fn cycle(&mut self) {
        self.0.cycle()
    }

    #[cfg(not(target_arch = "wasm32"))]
    pub fn run(&mut self) {
        self.0.run()
    }
}

pub struct SomeRom();
pub struct NoRom();
pub struct SomeWindow(Window);
pub struct NoWindow();

pub struct GeorgeEmuBuilder<Rom, Window> {
    rom: Rom,
    cpu: Option<Cpu>,
    keyboard: Option<Keyboard>,
    renderer: Option<Renderer>,
    cpu_controller: Option<CpuController>,
    memory_handle: Option<MemHandle>,
    window: Window,
}

impl GeorgeEmuBuilder<NoRom, NoWindow> {
    fn new() -> Self {
        GeorgeEmuBuilder {
            rom: NoRom(),
            cpu: None,
            keyboard: None,
            renderer: None,
            cpu_controller: None,
            memory_handle: None,
            window: NoWindow(),
        }
    }
    pub fn rom(self, rom: [u8; 0x8000]) -> GeorgeEmuBuilder<SomeRom, NoWindow> {
        let mut memory = Mem::new();
        memory.load_bytes(&rom);
        let memory_handle = MemHandle::new(memory);

        let (cpu, cpu_controller) = Cpu::new_with_control(memory_handle.clone());

        let Self {
            renderer,
            keyboard,
            window,
            ..
        } = self;

        GeorgeEmuBuilder {
            cpu: Some(cpu),
            rom: SomeRom(),
            keyboard,
            renderer,
            cpu_controller: Some(cpu_controller),
            memory_handle: Some(memory_handle),
            window,
        }
    }
}

impl GeorgeEmuBuilder<SomeRom, NoWindow> {
    /// Adds a minifb window to the emulator
    pub fn window(self, mut window: Window) -> GeorgeEmuBuilder<SomeRom, SomeWindow> {
        let Self {
            rom,
            cpu,
            keyboard,
            ref cpu_controller,
            ref memory_handle,
            ..
        } = self;
        let renderer = Renderer::new(
            self.cpu_controller.clone().unwrap(),
            self.memory_handle.clone().unwrap(),
        );

        window.set_input_callback(Box::new(Keyboard::new(
            self.memory_handle.as_ref().unwrap().clone(),
        )));

        GeorgeEmuBuilder {
            cpu,
            rom,
            keyboard,
            renderer: Some(renderer),
            cpu_controller: cpu_controller.clone(),
            memory_handle: memory_handle.clone(),
            window: SomeWindow(window),
        }
    }

    /// Enables debug printing and breakpoint triggering
    pub fn debug(mut self) -> Self {
        self.cpu.as_mut().unwrap().debug = true;
        self
    }

    #[cfg(not(target_arch = "wasm32"))]
    pub fn build(self) -> GeorgeEmu {
        let keyboard = Keyboard::new(self.memory_handle.clone().unwrap());
        GeorgeEmu(imp::GeorgeEmu::new(
            self.cpu.unwrap(),
            self.renderer.unwrap(),
            Some(keyboard),
            None,
        ))
    }
}

impl GeorgeEmuBuilder<NoRom, SomeWindow> {
    pub fn rom(self, rom: [u8; 0x8000]) -> GeorgeEmuBuilder<SomeRom, SomeWindow> {
        let mut memory = Mem::new();
        memory.load_bytes(&rom);
        let memory_handle = MemHandle::new(memory);

        let (cpu, cpu_controller) = Cpu::new_with_control(memory_handle.clone());

        let Self {
            renderer,
            keyboard,
            window,
            ..
        } = self;

        GeorgeEmuBuilder {
            cpu: Some(cpu),
            rom: SomeRom(),
            keyboard,
            renderer,
            cpu_controller: Some(cpu_controller),
            memory_handle: Some(memory_handle),
            window,
        }
    }

    /// Enables debug printing and breakpoint triggering
    pub fn debug(mut self) -> Self {
        self.cpu.as_mut().unwrap().debug = true;
        self
    }
}

impl GeorgeEmuBuilder<SomeRom, SomeWindow> {
    /// Enables debug printing and breakpoint triggering
    pub fn debug(mut self) -> Self {
        self.cpu.as_mut().unwrap().debug = true;
        self
    }

    #[cfg(not(target_arch = "wasm32"))]
    pub fn build(self) -> GeorgeEmu {
        let keyboard = Keyboard::new(self.memory_handle.clone().unwrap());
        GeorgeEmu(imp::GeorgeEmu::new(
            self.cpu.unwrap(),
            self.renderer.unwrap(),
            None,
            Some(self.window.0),
        ))
    }

    #[cfg(target_arch = "wasm32")]
    pub fn build(self) -> GeorgeEmu {
        GeorgeEmu(imp::GeorgeEmu::new(
            self.cpu.unwrap(),
            self.renderer.unwrap(),
            self.window.0,
        ))
    }
}
