use minifb::Window;
use std::{
    io::stdout,
    process::exit,
    thread::sleep,
    time::{Duration, Instant},
};
use termion::{
    async_stdin, clear,
    cursor::{self, Goto},
    event::Key,
    input::TermRead,
    raw::IntoRawMode,
};

use crate::{cpu::Cpu, keyboard::Keyboard, video::Renderer};

#[derive(Debug)]
pub struct GeorgeEmu {
    cpu: Cpu,
    renderer: Renderer,
    input: Option<Keyboard>,
    window: Option<Window>,
}

impl GeorgeEmu {
    pub fn new(
        mut cpu: Cpu,
        renderer: Renderer,
        input: Option<Keyboard>,
        window: Option<Window>,
    ) -> Self {
        cpu.reset();
        Self {
            cpu,
            input,
            renderer,
            window,
        }
    }

    pub fn draw(&mut self) {
        self.renderer.render(self.window.as_mut())
    }

    pub fn cycle(&mut self) {
        self.cpu.cycle()
    }

    pub fn run(&mut self) {
        if self.window.is_none() {
            let _ = stdout().into_raw_mode();
            print!("{}{}", cursor::Hide, clear::All,);
        }
        let clock_interval = Duration::from_nanos(500);
        let frame_interval = Duration::from_millis(16);
        let mut next_clock = Instant::now() + clock_interval;
        let mut next_frame = Instant::now() + frame_interval;

        loop {
            let now = Instant::now();

            if now >= next_clock {
                self.cpu.cycle();
                next_clock += clock_interval;
            }

            if now >= next_frame {
                if self.window.is_none() {
                    self.handle_input();
                }
                self.draw();
                next_frame += frame_interval;
            }

            let next_run = std::cmp::min(next_clock, next_frame);
            let sleep_dur = next_run - now;
            sleep(sleep_dur)
        }
    }

    fn handle_input(&mut self) {
        let mut stdin = async_stdin().keys();
        if let Some(Ok(key)) = stdin.next() {
            match key {
                Key::Char('q') => {
                    print!("{}{}{}", clear::All, cursor::Show, Goto(1, 1));
                    exit(0);
                }
                Key::Char('`') => self.cpu.toggle_stopped(),
                Key::Char('\n') => self.cpu.cycle(),
                Key::Char('i') => self.cpu.interrupt(),
                _ => {
                    self.input.as_mut().unwrap().read_keys(key);
                }
            }
        } else {
            self.input.as_mut().unwrap().clear_keys();
        }
    }
}
